
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "kgx-enums.h"
#include "kgx-close-dialog.h"
#include "kgx-colour-utils.h"
#include "kgx-paste-dialog.h"
#include "kgx-settings.h"
#include "kgx-spad.h"
#include "kgx-terminal.h"
#include "kgx-train.h"
#include "kgx-utils.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "kgx-close-dialog.h" */

GType
kgx_close_dialog_context_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (KGX_CONTEXT_WINDOW), "KGX_CONTEXT_WINDOW", "window" },
        { C_ENUM (KGX_CONTEXT_TAB), "KGX_CONTEXT_TAB", "tab" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("KgxCloseDialogContext"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
kgx_close_dialog_result_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (KGX_CLOSE_ANYWAY), "KGX_CLOSE_ANYWAY", "anyway" },
        { C_ENUM (KGX_CLOSE_CANCELLED), "KGX_CLOSE_CANCELLED", "cancelled" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("KgxCloseDialogResult"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "kgx-colour-utils.h" */

GType
kgx_colour_parse_error_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (KGX_COLOUR_PARSE_ERROR_WRONG_LENGTH), "KGX_COLOUR_PARSE_ERROR_WRONG_LENGTH", "colour-parse-error-wrong-length" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("KgxColourParseError"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "kgx-paste-dialog.h" */

GType
kgx_paste_dialog_result_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (KGX_PASTE_ANYWAY), "KGX_PASTE_ANYWAY", "anyway" },
        { C_ENUM (KGX_PASTE_CANCELLED), "KGX_PASTE_CANCELLED", "cancelled" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("KgxPasteDialogResult"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "kgx-settings.h" */

GType
kgx_theme_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (KGX_THEME_AUTO), "KGX_THEME_AUTO", "auto" },
        { C_ENUM (KGX_THEME_NIGHT), "KGX_THEME_NIGHT", "night" },
        { C_ENUM (KGX_THEME_DAY), "KGX_THEME_DAY", "day" },
        { C_ENUM (KGX_THEME_HACKER), "KGX_THEME_HACKER", "hacker" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("KgxTheme"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "kgx-spad.h" */

GType
kgx_spad_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (KGX_SPAD_NONE), "KGX_SPAD_NONE", "none" },
        { C_FLAGS (KGX_SPAD_SHOW_REPORT), "KGX_SPAD_SHOW_REPORT", "show-report" },
        { C_FLAGS (KGX_SPAD_SHOW_SYS_INFO), "KGX_SPAD_SHOW_SYS_INFO", "show-sys-info" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("KgxSpadFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "kgx-terminal.h" */

GType
kgx_zoom_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (KGX_ZOOM_IN), "KGX_ZOOM_IN", "in" },
        { C_ENUM (KGX_ZOOM_OUT), "KGX_ZOOM_OUT", "out" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("KgxZoom"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "kgx-train.h" */

GType
kgx_status_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (KGX_NONE), "KGX_NONE", "none" },
        { C_FLAGS (KGX_REMOTE), "KGX_REMOTE", "remote" },
        { C_FLAGS (KGX_PRIVILEGED), "KGX_PRIVILEGED", "privileged" },
        { C_FLAGS (KGX_PLAYBOX), "KGX_PLAYBOX", "playbox" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("KgxStatus"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "kgx-utils.h" */

GType
kgx_argument_error_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (KGX_ARGUMENT_ERROR_BOTH), "KGX_ARGUMENT_ERROR_BOTH", "argument-error-both" },
        { C_ENUM (KGX_ARGUMENT_ERROR_MISSING), "KGX_ARGUMENT_ERROR_MISSING", "argument-error-missing" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("KgxArgumentError"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

