/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* BOOLEAN: STRING, VARIANT (../gnome-console/src/kgx-marshals.list:1) */
G_GNUC_INTERNAL
void kgx_marshals_BOOLEAN__STRING_VARIANT (GClosure     *closure,
                                           GValue       *return_value,
                                           guint         n_param_values,
                                           const GValue *param_values,
                                           gpointer      invocation_hint,
                                           gpointer      marshal_data);
G_GNUC_INTERNAL
void kgx_marshals_BOOLEAN__STRING_VARIANTv (GClosure *closure,
                                            GValue   *return_value,
                                            gpointer  instance,
                                            va_list   args,
                                            gpointer  marshal_data,
                                            int       n_params,
                                            GType    *param_types);

/* BOXED: BOOLEAN, BOXED, BOXED (../gnome-console/src/kgx-marshals.list:2) */
G_GNUC_INTERNAL
void kgx_marshals_BOXED__BOOLEAN_BOXED_BOXED (GClosure     *closure,
                                              GValue       *return_value,
                                              guint         n_param_values,
                                              const GValue *param_values,
                                              gpointer      invocation_hint,
                                              gpointer      marshal_data);
G_GNUC_INTERNAL
void kgx_marshals_BOXED__BOOLEAN_BOXED_BOXEDv (GClosure *closure,
                                               GValue   *return_value,
                                               gpointer  instance,
                                               va_list   args,
                                               gpointer  marshal_data,
                                               int       n_params,
                                               GType    *param_types);

/* INT64: BOOLEAN, INT64 (../gnome-console/src/kgx-marshals.list:3) */
G_GNUC_INTERNAL
void kgx_marshals_INT64__BOOLEAN_INT64 (GClosure     *closure,
                                        GValue       *return_value,
                                        guint         n_param_values,
                                        const GValue *param_values,
                                        gpointer      invocation_hint,
                                        gpointer      marshal_data);
G_GNUC_INTERNAL
void kgx_marshals_INT64__BOOLEAN_INT64v (GClosure *closure,
                                         GValue   *return_value,
                                         gpointer  instance,
                                         va_list   args,
                                         gpointer  marshal_data,
                                         int       n_params,
                                         GType    *param_types);

/* OBJECT: VOID (../gnome-console/src/kgx-marshals.list:4) */
G_GNUC_INTERNAL
void kgx_marshals_OBJECT__VOID (GClosure     *closure,
                                GValue       *return_value,
                                guint         n_param_values,
                                const GValue *param_values,
                                gpointer      invocation_hint,
                                gpointer      marshal_data);
G_GNUC_INTERNAL
void kgx_marshals_OBJECT__VOIDv (GClosure *closure,
                                 GValue   *return_value,
                                 gpointer  instance,
                                 va_list   args,
                                 gpointer  marshal_data,
                                 int       n_params,
                                 GType    *param_types);

/* VOID: BOXED (../gnome-console/src/kgx-marshals.list:5) */
#define kgx_marshals_VOID__BOXED	g_cclosure_marshal_VOID__BOXED
#define kgx_marshals_VOID__BOXEDv	g_cclosure_marshal_VOID__BOXEDv

/* VOID: ENUM (../gnome-console/src/kgx-marshals.list:6) */
#define kgx_marshals_VOID__ENUM	g_cclosure_marshal_VOID__ENUM
#define kgx_marshals_VOID__ENUMv	g_cclosure_marshal_VOID__ENUMv

/* VOID: ENUM, STRING, BOOLEAN (../gnome-console/src/kgx-marshals.list:7) */
G_GNUC_INTERNAL
void kgx_marshals_VOID__ENUM_STRING_BOOLEAN (GClosure     *closure,
                                             GValue       *return_value,
                                             guint         n_param_values,
                                             const GValue *param_values,
                                             gpointer      invocation_hint,
                                             gpointer      marshal_data);
G_GNUC_INTERNAL
void kgx_marshals_VOID__ENUM_STRING_BOOLEANv (GClosure *closure,
                                              GValue   *return_value,
                                              gpointer  instance,
                                              va_list   args,
                                              gpointer  marshal_data,
                                              int       n_params,
                                              GType    *param_types);

/* VOID: INT (../gnome-console/src/kgx-marshals.list:8) */
#define kgx_marshals_VOID__INT	g_cclosure_marshal_VOID__INT
#define kgx_marshals_VOID__INTv	g_cclosure_marshal_VOID__INTv

/* VOID: STRING (../gnome-console/src/kgx-marshals.list:9) */
#define kgx_marshals_VOID__STRING	g_cclosure_marshal_VOID__STRING
#define kgx_marshals_VOID__STRINGv	g_cclosure_marshal_VOID__STRINGv

/* VOID: UINT, UINT (../gnome-console/src/kgx-marshals.list:10) */
G_GNUC_INTERNAL
void kgx_marshals_VOID__UINT_UINT (GClosure     *closure,
                                   GValue       *return_value,
                                   guint         n_param_values,
                                   const GValue *param_values,
                                   gpointer      invocation_hint,
                                   gpointer      marshal_data);
G_GNUC_INTERNAL
void kgx_marshals_VOID__UINT_UINTv (GClosure *closure,
                                    GValue   *return_value,
                                    gpointer  instance,
                                    va_list   args,
                                    gpointer  marshal_data,
                                    int       n_params,
                                    GType    *param_types);

/* VOID: VOID (../gnome-console/src/kgx-marshals.list:11) */
#define kgx_marshals_VOID__VOID	g_cclosure_marshal_VOID__VOID
#define kgx_marshals_VOID__VOIDv	g_cclosure_marshal_VOID__VOIDv


G_END_DECLS
