
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "kgx-close-dialog.h" */

GType kgx_close_dialog_context_get_type (void);
#define KGX_TYPE_CLOSE_DIALOG_CONTEXT (kgx_close_dialog_context_get_type())

GType kgx_close_dialog_result_get_type (void);
#define KGX_TYPE_CLOSE_DIALOG_RESULT (kgx_close_dialog_result_get_type())

/* enumerations from "kgx-colour-utils.h" */

GType kgx_colour_parse_error_get_type (void);
#define KGX_TYPE_COLOUR_PARSE_ERROR (kgx_colour_parse_error_get_type())

/* enumerations from "kgx-paste-dialog.h" */

GType kgx_paste_dialog_result_get_type (void);
#define KGX_TYPE_PASTE_DIALOG_RESULT (kgx_paste_dialog_result_get_type())

/* enumerations from "kgx-settings.h" */

GType kgx_theme_get_type (void);
#define KGX_TYPE_THEME (kgx_theme_get_type())

/* enumerations from "kgx-spad.h" */

GType kgx_spad_flags_get_type (void);
#define KGX_TYPE_SPAD_FLAGS (kgx_spad_flags_get_type())

/* enumerations from "kgx-terminal.h" */

GType kgx_zoom_get_type (void);
#define KGX_TYPE_ZOOM (kgx_zoom_get_type())

/* enumerations from "kgx-train.h" */

GType kgx_status_get_type (void);
#define KGX_TYPE_STATUS (kgx_status_get_type())

/* enumerations from "kgx-utils.h" */

GType kgx_argument_error_get_type (void);
#define KGX_TYPE_ARGUMENT_ERROR (kgx_argument_error_get_type())

G_END_DECLS

/* Generated data ends here */

